"""
While real functions don't exist in the eiko language,
constructors and plugins do, and they need some kind of representation.
"""
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Type, Union

from ...errors import EikoCompilationError, EikoInternalError, EikoPluginError
from ...plugin import EikoPluginException, EikoPluginTyping
from .._token import Token
from .base_types import (
    INDEXABLE_TYPES,
    EikoBaseType,
    EikoProtectedStr,
    EikoResource,
    EikoType,
    EikoUnset,
    PassedArg,
    PyTypes,
    to_eiko,
    to_eiko_type,
)

if TYPE_CHECKING:
    from .._parser import ExprAST
    from ._resource import EikoResourceDefinition
    from .context import CompilerContext, StorableTypes

EikoFunctionType = EikoType("function")


@dataclass
class ConstructorArg:
    """Representation of a required constructor argument."""

    name: str
    type: EikoType
    default_value: Optional[EikoBaseType] = None


class ConstructorDefinition(EikoBaseType):
    """Internal representation of an Eikobot constructor."""

    def __init__(self, name: str, execution_context: "CompilerContext") -> None:
        super().__init__(EikoFunctionType)
        self.parent: "EikoResourceDefinition"
        self.name = name
        self.args: dict[str, ConstructorArg] = {}
        self.body: list["ExprAST"] = []
        self.execution_context = execution_context
        self.index_def: list[str] = []

    def printable(self, _: str = "") -> str:
        raise NotImplementedError

    def add_arg(self, arg: ConstructorArg) -> None:
        self.args[arg.name] = arg

    def add_body_expr(self, expr: "ExprAST") -> None:
        self.body.append(expr)

    def execute(  # pylint: disable=too-many-locals,too-many-branches
        self,
        callee_token: Token,
        positional_args: list[PassedArg],
        keyword_args: dict[str, PassedArg],
    ) -> EikoResource:
        """
        Executes a function call based on a premade constructor spec.
        """
        if self.parent.promises and self.parent.handler is None:
            raise EikoCompilationError(
                f"Resource Definition '{self.parent.name}' has promises, "
                "but no handler",
                token=self.parent.expr.token,
            )

        if len(positional_args) + len(keyword_args) > len(self.args) - 1:
            raise EikoCompilationError(
                "Too many arguments given to function call. "
                f"Expected {len(self.args) - 1}, "
                f"got {len(positional_args) + len(keyword_args)}.",
                token=callee_token,
            )

        handled_args: dict[str, EikoBaseType] = {}
        self_arg = list(self.args.values())[0]
        resource = EikoResource(self.parent)
        handled_args[self_arg.name] = resource
        self._handle_args(handled_args, positional_args, keyword_args)

        for arg_name, arg in self.args.items():
            if arg_name not in handled_args:
                if arg.default_value is None:
                    raise EikoCompilationError(
                        f"Argument '{arg_name}' for callable '{callee_token.content}' requires a value.",
                        token=callee_token,
                    )

                handled_args[arg_name] = arg.default_value

        context = self.execution_context.get_subcontext(
            f"{self.parent.name}.{self.name}"
        )
        for prop in self.parent.properties.values():
            if prop in self.parent.promises:
                # prop is garantueed to be an EikoPromiseDef here
                resource.add_promise(prop.execute(callee_token, resource))  # type: ignore
            else:
                resource.populate_property(prop.name, prop.type)

        for arg_name, value in handled_args.items():
            context.set(arg_name, value)

        for expr in self.body:
            expr.compile(context)

        for prop in self.parent.properties.values():
            res_prop = resource.properties.get(prop.name)
            if isinstance(res_prop, EikoUnset):
                raise EikoCompilationError(
                    f"Property '{self.parent.name}.{prop.name}' was not set in the constructor.",
                    token=callee_token,
                )

        res_index = self.parent.name
        for property_name in self.index_def:
            if property_name == self.parent.name:
                continue
            prop_name_split = property_name.split(".")
            index_prop: Union[EikoBaseType, EikoUnset, None] = resource
            for prop_name in prop_name_split:
                if isinstance(index_prop, EikoResource):
                    index_prop = index_prop.properties.get(prop_name)
                else:
                    raise EikoCompilationError(
                        f"Failed to create index using property '{property_name}': "
                        f"No property '{prop_name}'.",
                        token=self.parent.expr.token,
                    )

            if not isinstance(index_prop, INDEXABLE_TYPES):
                raise EikoCompilationError(
                    f"Property '{property_name}' of '{self.parent.name}' is not an indexable type.",
                    token=self.parent.expr.token,
                )

            if isinstance(index_prop, EikoProtectedStr):
                raise EikoCompilationError(
                    f"Property '{property_name}' of '{self.parent.name}' is a protected string. "
                    "It can not be used to create an index.",
                    token=self.parent.expr.token,
                )

            res_index += "-" + index_prop.index()

        resource.set_index(res_index)
        if resource.index() in context.global_id_list:
            raise EikoCompilationError(
                f"A resource of type '{self.parent.name}' with index "
                f"'{resource.index()}' was already created.",
                token=callee_token,
            )

        context.global_id_list.append(resource.index())

        return resource

    def _handle_args(  # pylint: disable=too-many-branches
        self,
        handled_args: dict[str, EikoBaseType],
        positional_args: list[PassedArg],
        keyword_args: dict[str, PassedArg],
    ) -> None:
        for passed_arg, arg in zip(positional_args, list(self.args.values())[1:]):
            if not passed_arg.value.type_check(arg.type):
                # Try to coerce the type
                if arg.type.inverse_type_check(passed_arg.value.type):
                    if arg.type.typedef is None:
                        raise EikoInternalError(
                            "Failed to coerce type.",
                            token=passed_arg.token,
                        )
                    passed_arg.value = arg.type.typedef.execute(
                        passed_arg.value, passed_arg.token
                    )
                else:
                    raise EikoCompilationError(
                        f"Argument '{arg.name}' expected value of type '{arg.type}', "
                        f"but got value of type '{passed_arg.value.type}'.",
                        token=passed_arg.token,
                    )

            handled_args[arg.name] = passed_arg.value

        for arg_name, passed_arg in keyword_args.items():
            positional_value = handled_args.get(arg_name)
            if positional_value is not None:
                raise EikoCompilationError(
                    f"'{self.name}()' got multiple values for argument '{arg_name}'."
                )

            kw_arg = self.args.get(arg_name)
            if kw_arg is None:
                raise EikoCompilationError(
                    f"'{self.parent.name}.{self.name}()' has no argument '{arg_name}'.",
                    token=passed_arg.token,
                )

            if not passed_arg.value.type_check(kw_arg.type):
                if kw_arg.type.inverse_type_check(passed_arg.value.type):
                    if kw_arg.type.typedef is not None:
                        passed_arg.value = kw_arg.type.typedef.execute(
                            passed_arg.value,
                            passed_arg.token,
                        )
                    else:
                        raise EikoInternalError(
                            "Ran in to a typedef that does not have a constructor. "
                            "This is most likely a bug. PLease report this on Github.",
                            token=passed_arg.token,
                        )
                else:
                    raise EikoCompilationError(
                        f"Argument '{kw_arg.name}' expected value of type '{kw_arg.type}', "
                        f"but got value of type '{passed_arg.value.type}'.",
                        token=passed_arg.token,
                    )

            handled_args[arg_name] = passed_arg.value

    def truthiness(self) -> bool:
        raise NotImplementedError


EikoPluginType = EikoType("plugin")


class DefaultValueNotSet:
    """Indicates no default value was set."""

    def __init__(self) -> None:
        pass


@dataclass
class PluginArg:
    """Class representing an argument in a plugin call."""

    name: str
    py_type: Union[Type[EikoBaseType], Type[PyTypes]]
    default_value: Union[PyTypes, DefaultValueNotSet] = DefaultValueNotSet()


class PluginDefinition(EikoBaseType):
    """
    Internal representation of a python plugin
    that can be called from the Eikobot language.
    """

    def __init__(
        self,
        body: EikoPluginTyping,
        return_type: Type[EikoBaseType],
        identifier: str,
        module: str,
    ) -> None:
        super().__init__(EikoPluginType)
        self.body = body
        self.return_type = to_eiko_type(return_type)
        self._body_return_type = return_type
        self.args: list[PluginArg] = []
        self.identifier = identifier
        self.module = module

    def printable(self, _: str = "") -> str:
        return f"Plugin '{self.identifier}'"

    def add_arg(self, arg: PluginArg) -> None:
        self.args.append(arg)

    def execute(
        self, args: list["ExprAST"], context: "CompilerContext", token: Optional[Token]
    ) -> Optional[EikoBaseType]:
        """Execute the stored function and coerces types."""

        stable_args = []
        for i, arg in enumerate(args):
            stable_args.append(self._handle_arg(arg, context, self.args[i]))

        try:
            val = self.body(*stable_args)
        except EikoPluginException as e:
            raise EikoPluginError(str(e), token=token) from e
        except Exception as e:
            raise EikoPluginError(
                f"An unhandled exception occured while executing plugin '{self.identifier}'",
                token=token,
                python_exception=e,
            ) from e

        return to_eiko(val)

    def _handle_arg(
        self, arg: "ExprAST", context: "CompilerContext", required_arg: PluginArg
    ) -> Union[EikoBaseType, bool, float, int, str]:
        compiled_arg = arg.compile(context)
        if compiled_arg is None:
            raise EikoCompilationError(
                f"Plugin '{self.module}.{self.identifier}' arg '{required_arg.name}' expects a value "
                f"but expression did not result in a suitable value.",
                token=arg.token,
            )
        if issubclass(required_arg.py_type, EikoBaseType):
            converted_arg: Union["StorableTypes", PyTypes] = compiled_arg
        else:
            if isinstance(compiled_arg, EikoBaseType):
                converted_arg = compiled_arg.to_py()
            else:
                raise EikoCompilationError(
                    "Failed to convert to python type when passing to plugin.",
                    token=arg.token,
                )

        if not isinstance(converted_arg, required_arg.py_type):
            raise EikoCompilationError(
                f"Plugin '{self.module}.{self.identifier}' arg '{required_arg.name}' expects an argument "
                f"of type '{required_arg.py_type.__name__}', but instead got '{compiled_arg.type}'.",
                token=arg.token,
            )

        return converted_arg  # type: ignore

    def truthiness(self) -> bool:
        raise NotImplementedError



"""
For help with creating, distributing and installing packages.
"""
import asyncio
import copy
import os
import shutil
import subprocess
import sys
import tarfile
import tomllib
from dataclasses import dataclass
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import aiohttp
from packaging import version
from pydantic import BaseModel, ValidationError

from ... import VERSION
from .. import logger
from ..compiler.importlib import INTERNAL_LIB_PATH

CACHE_PATH = Path(__file__).parent / "cache"
CACHE_PATH.mkdir(exist_ok=True)
LIB_PATH = Path(__file__).parent / "lib"
LIB_PATH.mkdir(exist_ok=True)

PKG_INDEX: dict[str, "PackageData"] = {}


class EikoPackageError(Exception):
    """An error that occured during the eiko compilation process."""

    def __init__(self, reason: str, *args: object) -> None:
        super().__init__(reason, *args)


class PackageData(BaseModel):
    """
    All the settings one might
    find in an eiko.toml file.
    """

    name: str
    source_dir: Path
    version: Optional[str] = None
    description: Optional[str] = None
    long_description: Optional[str] = None
    long_description_content_type: Optional[str] = None
    author: Optional[str] = None
    author_email: Optional[str] = None
    license: Optional[str] = None
    eikobot_requires: Optional[str] = None
    requires: list[str] = []

    def pkg_name_version(self) -> str:
        """
        Returns the name that was used for the install dir
        under package_manager/lib.
        """
        name = self.name
        if self.version is not None:
            name += "-" + self.version

        return name

    def eiko_version_match(self) -> bool:  # pylint: disable=too-many-branches
        """
        Checks to see if the library is compatible with
        the currently installed version of eikobot.
        """
        if self.eikobot_requires is None:
            return True

        eiko_version = version.parse(VERSION)
        for version_req in self.eikobot_requires.split(","):
            if version_req.startswith(">="):
                if eiko_version < _parse_version(version_req[2:]):
                    return False

            elif version_req.startswith("<="):
                if eiko_version > _parse_version(version_req[2:]):
                    return False

            elif version_req.startswith("=="):
                if eiko_version != _parse_version(version_req[2:]):
                    return False

            elif version_req.startswith("!="):
                if eiko_version == _parse_version(version_req[2:]):
                    return False

            elif version_req.startswith(">"):
                if eiko_version <= _parse_version(version_req[1:]):
                    return False

            elif version_req.startswith("<"):
                if eiko_version >= _parse_version(version_req[1:]):
                    return False

            else:
                raise EikoPackageError(
                    f"Failed to parse eiko.toml option eikobot_requires '{self.eikobot_requires}'."
                )

        return True


def _parse_version(_version: str) -> version.Version:
    try:
        return version.parse(_version)
    except version.InvalidVersion as e:
        raise EikoPackageError(
            f"failed to parse `eikobot_requires` version `{_version}`"
        ) from e


def _construct_pkg_index() -> None:
    for pkg_dir in LIB_PATH.glob("*"):
        pkg_data = _read_pkg_toml(pkg_dir / "eiko.toml")
        PKG_INDEX[pkg_data.name] = pkg_data


def _read_pkg_toml(path: Path) -> PackageData:
    logger.debug("Reading eiko toml file.")
    toml = tomllib.loads(path.read_text(encoding="utf-8"))
    pkg_toml = toml.get("eiko", {}).get("package")
    if pkg_toml is None:
        raise EikoPackageError("eiko.toml does not contain an 'eiko.package' section.")
    try:
        pkg_data = PackageData(**pkg_toml)
    except ValidationError as e:
        raise EikoPackageError(f"Failed to parse eiko.toml: {e}") from e

    return pkg_data


def build_pkg() -> None:
    """Builds a package in the cwd"""
    logger.info("Building package.")
    eiko_toml_path = Path("eiko.toml")
    if not eiko_toml_path.exists():
        raise EikoPackageError("eiko.toml file missing.")

    pkg_data = _read_pkg_toml(eiko_toml_path)
    dist = Path("dist")
    dist.mkdir(exist_ok=True)
    if not pkg_data.source_dir.exists():
        raise EikoPackageError(f"Not such source directory: '{pkg_data.source_dir}'.")

    if not pkg_data.source_dir.is_dir():
        raise EikoPackageError(f"'{pkg_data.source_dir}' is not a directory.")

    logger.debug("Setting up build env.")
    build_dir = dist / "build"
    shutil.rmtree(build_dir, ignore_errors=True)
    build_dir.mkdir()
    # Compile every file here to make sure they are valid?
    # This would also be the place to check for dangerous sources.
    shutil.copytree(
        pkg_data.source_dir,
        build_dir / pkg_data.source_dir,
        ignore=shutil.ignore_patterns("__pycache__"),
    )
    shutil.copy(eiko_toml_path, build_dir)

    logger.debug("Creating tar archive.")
    dist_name = pkg_data.pkg_name_version()
    dist_file_name = dist_name + ".eiko.tar.gz"
    dist_file = dist / dist_file_name

    with tarfile.open(dist_file, "w:gz") as archive:
        archive.add(build_dir, arcname=dist_name)

    logger.debug("cleaning up build env.")
    shutil.rmtree(build_dir, ignore_errors=True)

    logger.info(f"Build package '{dist_name}'")


def install_pkg(pkg_def: str) -> None:
    """
    This functions gets a string prompt and will try to figure out
    if the package is http(s), git or a path.
    """
    _construct_pkg_index()
    if pkg_def.startswith("http://") or pkg_def.startswith("https://"):
        _download_pkg(pkg_def)

    elif pkg_def.endswith(".eiko.tar.gz"):
        _install_pkg_from_path(Path(pkg_def))

    else:
        raise EikoPackageError("Failed to properly parse package url or path.")


def _download_pkg(url: str) -> None:
    if url.endswith(".eiko.tar.gz"):
        logger.debug("Directly downloading package.")
        asyncio.run(_download_to_cache(url))

    else:
        raise EikoPackageError(
            "Currently only direct links to an eiko package are supported."
        )


def _human_readable(number: int) -> str:
    number = number // 8
    for unit in ["", "K", "M", "G", "T"]:
        if number < 10240:
            return f"{number}{unit}B"
        number = number // 1024
    return f"{number}PB"


async def _download_to_cache(url: str) -> None:
    pkg_path = CACHE_PATH / Path(urlparse(url).path).name
    async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(10)) as session:
        logger.info(f"Fetching {url}")
        try:
            response = await session.get(url)
            total_dl_size = int(response.headers.get("CONTENT-LENGTH", 0))
            downloaded = 0

            with open(pkg_path, "wb") as f:
                async for data, _ in response.content.iter_chunks():
                    f.write(data)
                    downloaded += sys.getsizeof(data)
                    percent = int((downloaded / total_dl_size) * 20)
                    pg_bar = "=" * percent
                    if percent == 20:
                        print(
                            f"    [{pg_bar}]",
                            f"{_human_readable(total_dl_size)}/{_human_readable(total_dl_size)}",
                        )
                    else:
                        print(
                            f"    [{pg_bar}>{(20 - percent - 1) * ' '}]",
                            f"{_human_readable(downloaded)}/{_human_readable(total_dl_size)}",
                            end="\r",
                        )

        except aiohttp.ClientError as e:
            raise EikoPackageError(f"Failed to download {url}: {e}") from e

    _install_pkg_from_cache(pkg_path.name)


def _install_pkg_from_path(pkg_path: Path) -> None:
    """
    Install a package using a path to an .eiko.tar.gz file.
    """
    if not pkg_path.exists():
        raise EikoPackageError(f"Package path does not exist: '{pkg_path}'.")
    logger.debug("Adding archive to cache.")
    shutil.copy(pkg_path, CACHE_PATH)
    _install_pkg_from_cache(pkg_path.name)


def _install_pkg_from_cache(archive_name: str) -> None:
    pkg_name = archive_name.removesuffix(".eiko.tar.gz")
    logger.debug("Unpacking archive.")
    with tarfile.open(CACHE_PATH / archive_name, "r:gz") as archive:
        archive.extractall(LIB_PATH)

    pkg_lib_path = LIB_PATH / pkg_name
    pkg_data = _read_pkg_toml(pkg_lib_path / "eiko.toml")
    prev_pkg = PKG_INDEX.get(pkg_data.name)
    if prev_pkg is not None:
        _uninstall_pkg(prev_pkg)

    if pkg_data.version is None:
        logger.debug(f"Installing '{pkg_data.name}'.")
    else:
        logger.debug(f"Installing '{pkg_data.name}=={pkg_data.version}'.")

    requirements_file = pkg_lib_path / "requirements.txt"
    if requirements_file.exists():
        logger.debug("Installing python requirements.")
        subprocess.run(
            [sys.executable, "-m", "pip", "install", "-r", requirements_file],
            check=True,
        )

    logger.debug("Installing requirements.")
    for req in pkg_data.requires:
        install_pkg(req)

    try:
        os.symlink(
            pkg_lib_path / pkg_data.source_dir, INTERNAL_LIB_PATH / pkg_data.source_dir
        )
    except FileExistsError:
        # This is a bug in the package index
        # I have no idea what causes it, but this is hack around the issue.
        os.remove(INTERNAL_LIB_PATH / pkg_data.source_dir)
        os.symlink(
            pkg_lib_path / pkg_data.source_dir, INTERNAL_LIB_PATH / pkg_data.source_dir
        )

    if pkg_data.version is None:
        logger.info(f"Installed '{pkg_data.name}'.")
    else:
        logger.info(f"Installed '{pkg_data.name}=={pkg_data.version}'.")


def get_installed_pkg() -> dict[str, PackageData]:
    _construct_pkg_index()
    return copy.deepcopy(PKG_INDEX)


def uninstall_pkg(name: str) -> None:
    """
    Uninstalls a package.
    """
    _construct_pkg_index()
    pkg_data = PKG_INDEX.get(name)
    if pkg_data is None:
        logger.error(f"Failed to install package: '{name}'")
        return

    _uninstall_pkg(pkg_data)


def _uninstall_pkg(pkg_data: PackageData) -> None:
    logger.debug(f"Uninstalling '{pkg_data.pkg_name_version()}'")
    os.remove(INTERNAL_LIB_PATH / pkg_data.source_dir)
    shutil.rmtree(LIB_PATH / pkg_data.pkg_name_version())
    logger.info(f"Uninstalled '{pkg_data.pkg_name_version()}'")
